/*
	Copyright (C) 2004 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/

//module for interfacing D with the win32 api!


module dwindows;

import win;
private import std.string;


int windowTextLength(HWND hwnd)
{
	return GetWindowTextLength(hwnd);
}


char[] windowText(HWND hwnd)
{
	int len = windowTextLength(hwnd);
	
	if(!len)
		return null;
	
	char[] s = new char[++len];
	GetWindowText(hwnd, s, len);
	return s[0 .. len - 1];
}


void windowText(HWND hwnd, char[] newText)
{
	SetWindowText(hwnd, toStringz(newText));
}


void displayWindow(HWND hwnd) //if the window is hidden it's shown, if it's minimized it's restored, if it's not active it's activated
{
	DWORD style;
	
	style = GetWindowLong(hwnd, GWL_STYLE);
	
	if(!(style & WS_VISIBLE))
		ShowWindow(hwnd, SW_SHOW);
	if(style & WS_MINIMIZE)
		ShowWindow(hwnd, SW_RESTORE);
	
	//SetForegroundWindow(hwnd);
	SetActiveWindow(hwnd);
}


void parentArea(HWND hwnd, RECT* area) //fills -area- with the area of -hwnd-'s parent; if -hwnd- has no parent, it uses the desktop working area
{
	HWND hwParent = GetParent(hwnd);
	
	if(hwParent)
	{
		if(GetWindowLong(hwnd, GWL_STYLE) & WS_CHILD)
			GetClientRect(hwParent, area);
		else
			GetWindowRect(hwParent, area);
	}
	else
	{
		SystemParametersInfo(SPI_GETWORKAREA, 0, area, FALSE);
	}
}


void centerWindow(HWND hwnd)
{
	RECT r, pr;
	HWND hwParent;
	
	parentArea(hwnd, &pr);
	
	GetWindowRect(hwnd, &r);
	hwParent = GetParent(hwnd);
	if(hwParent)
	{
		if(GetWindowLong(hwnd, GWL_STYLE) & WS_CHILD)
			MapWindowPoints(null, hwParent, (POINT*)&r, 2);
	}
	
	SetWindowPos(hwnd, null, pr.left + (((pr.right - pr.left) - (r.right - r.left)) / 2), pr.top + (((pr.bottom - pr.top) - (r.bottom - r.top)) / 2), 0, 0, SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);
}


enum Align
{
	LEFT =		0x01,
	TOP =		0x02,
	RIGHT =		0x04,
	BOTTOM =	0x08,
}


void alignWindow(HWND hwnd, Align a)
{
	int num;
	RECT r, pr;
	HWND hwParent;
	
	parentArea(hwnd, &pr);
	
	GetWindowRect(hwnd, &r);
	hwParent = GetParent(hwnd);
	if(hwParent)
	{
		if(GetWindowLong(hwnd, GWL_STYLE) & WS_CHILD)
			MapWindowPoints(null, hwParent, (POINT*)&r, 2);
	}
	
	if(a & Align.LEFT)
	{
		if(a & Align.RIGHT)
		{
			r.left = pr.left;
			r.right = pr.right;
		}
		else
		{
			num = r.right - r.left;
			r.left = pr.left;
			r.right = r.left + num;
		}
	}
	else
	{
		if(a & Align.RIGHT)
		{
			num = r.right - r.left;
			r.right = pr.right;
			r.left = r.right - num;
		}
	}
	if(a & Align.TOP)
	{
		if(a & Align.BOTTOM)
		{
			r.top = pr.top;
			r.bottom = pr.bottom;
		}
		else
		{
			num = r.bottom - r.top;
			r.top = pr.top;
			r.bottom = r.top + num;
		}
	}
	else
	{
		if(a & Align.BOTTOM)
		{
			num = r.bottom - r.top;
			r.bottom = pr.bottom;
			r.top = r.bottom - num;
		}
	}
	
	SetWindowPos(hwnd, null, r.left, r.top, 0, 0, SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOSIZE);
}


const char[] DIRTOOLONG = "directory too long";


char[] windowsDirectory()
{
	char[] s = new char[MAX_PATH];
	UINT rval;
	
	rval = GetWindowsDirectory(s, s.length);
	if(rval > s.length)
		throw new Exception(DIRTOOLONG);
	return s[0 .. rval];
}


char[] currentDirectory()
{
	char[] s = new char[MAX_PATH];
	UINT rval;
	
	rval = GetCurrentDirectory(s.length, s);
	if(rval > s.length)
		throw new Exception(DIRTOOLONG);
	return s[0 .. rval];
}


enum WndClassFlags: uint //which members of a WNDCLASSEX struct are valid (WNDCLASSEX.lpszClassName and WNDCLASSEX.hInstance are always valid); WNDCLASSEX.cbSize need not be valid
{
	NONE =			0x00000000, //must not be combined with others
	
	STYLE =			0x00000001, //WNDCLASSEX.style is valid
	ICONS =			0x00000002, //WNDCLASSEX.hIcon and WNDCLASSEX.hIconSm are valid
	CURSOR = 		0x00000004, //WNDCLASSEX.hCursor is valid
	BACKGROUND =	0x00000008, //WNDCLASSEX.hbrBackground is valid
	MENU = 			0x00000010, //WNDCLASSEX.lpszMenuName is valid
	
	ALL =			0x0000001F, //all of the above
}


class WindowException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}


class RegisterWindowException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


int messageBox(HWND hwParent, char[] msg, char[] title, UINT type)
{
	return MessageBox(hwParent, toStringz(msg), toStringz(title), type);
}


int messageBox(HWND hwParent, char[] msg, char[] title)
{
	return MessageBox(hwParent, toStringz(msg), toStringz(title), MB_OK);
}


int messageBox(HWND hwParent, char[] msg)
{
	return MessageBox(hwParent, toStringz(msg), "", MB_OK);
}


int messageLoop()
{
	MSG msg;
	while(GetMessage(&msg, null, 0, 0))
	{
		if(!IsDialogMessage(GetActiveWindow(), &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return cast(int)msg.wParam;
}


//doesn't disable the parent because it can't enable it before
//hwnd is hidden; resulting in another window activating
int showModal(HWND hwnd, int nShow)
{
	MSG msg;
	ShowWindow(hwnd, nShow);
	while(GetWindowLong(hwnd, GWL_STYLE) & WS_VISIBLE)
	{
		if(!GetMessage(&msg, null, 0, 0))
		{
			PostQuitMessage(msg.wParam);
			return cast(int)msg.wParam;
		}
		
		if(!IsDialogMessage(GetActiveWindow(), &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return 0;
}


class Window
{
	static Window[] windows;
	
	
	static int findIndex(HWND hwnd) //finds the index of -hwnd- and retruens the index in -windows-, -1 if not found
	{
		int i;
		
		for(i = 0; i != windows.length; i++)
		{
			if(windows[i]._hwnd == hwnd)
				return i;
		}
		
		return -1;
	}
	
	
	static Window find(HWND hwnd)
	{
		int i = findIndex(hwnd);
		return (i == -1) ? null : windows[i];
	}
	
	
	static int findIndex(Window w)
	{
		int i;
		
		for(i = 0; i != windows.length; i++)
		{
			if(windows[i] === w)
				return i;
		}
		
		return -1;
	}
	
	
	private static void remove(int index)
	{
		if(index == windows.length - 1)
			windows = windows[0 .. windows.length - 1];
		else
			windows = windows[0 .. index] ~ windows[index + 1 .. windows.length];
	}
	
	
	private void addThis()
	{
		windows.length = windows.length + 1;
		windows[windows.length - 1] = this;
	}
	
	
	private void removeThis()
	{
		int i = findIndex(_hwnd);
		if(i == -1)
			throw new WindowException("Window not found");
		remove(i);
	}
	
	
	HWND _hwnd = null;
	WNDPROC _prevWindowProc = null;
	
	
	LRESULT prevWindowProc(UINT msg, WPARAM wparam, LPARAM lparam) //used when subclassing a window; should be called and returned if a message isn't processed instead of super.windowProc
	{
		return CallWindowProc(_prevWindowProc, _hwnd, msg, wparam, lparam);
	}
	
	
	int isSubClass()
	{
		return cast(int)&_prevWindowProc;
	}
	
	
	int findThisIndex()
	{
		return findIndex(this);
	}
	
	
	HWND hwnd()
	{
		return _hwnd;
	}
	
	
	static WNDPROC superClass(HINSTANCE hinst, char[] className, WNDCLASSEX* newClassInfo, WndClassFlags flags) //-flags- indicates which members of -newClassInfo- are to be updated; returns the previous window procedure and should be called and returned if a message isn't processed instead of super.windowProc
	{
		WNDCLASSEX info;
		WNDPROC wndProc;
		
		info.cbSize = WNDCLASSEX.size;
		if(!GetClassInfoEx(hinst, toStringz(className), &info))
			throw new RegisterWindowException("Unable to obtain information for window class '" ~ className ~ "'");
		
		wndProc = info.lpfnWndProc;
		
		//optinal members
		if(flags & WndClassFlags.STYLE)
			info.style = newClassInfo.style;
		if(flags & WndClassFlags.ICONS)
		{
			info.hIcon = newClassInfo.hIcon;
			info.hIconSm = newClassInfo.hIconSm;
		}
		if(flags & WndClassFlags.CURSOR)
			info.hCursor = newClassInfo.hCursor;
		if(flags & WndClassFlags.BACKGROUND)
			info.hbrBackground = newClassInfo.hbrBackground;
		if(flags & WndClassFlags.MENU)
			info.lpszMenuName = newClassInfo.lpszMenuName;
		
		//required members
		info.lpszClassName = newClassInfo.lpszClassName;
		info.hInstance = newClassInfo.hInstance;
		
		info.lpfnWndProc = (WNDPROC)&allWindowProc;
		
		registerClass(&info);
		return wndProc;
	}
	
	
	void subClass(HWND hwnd)
	{
		addThis();
		_hwnd = hwnd;
		_prevWindowProc = (WNDPROC)SetWindowLong(hwnd, GWL_WNDPROC, (LONG)&allWindowProc);
		if(!_prevWindowProc)
		{
			_hwnd = null;
			removeThis();
			throw new WindowException("Unable to subclass window");
		}
	}
	
	
	static void registerClass(WNDCLASSEX* wce)
	{
		wce.lpfnWndProc = (WNDPROC)&allWindowProc;
		wce.cbSize = WNDCLASSEX.size;
		wce.cbClsExtra = 0;
		wce.cbWndExtra = 0;
		
		if(!RegisterClassEx(wce))
			throw new RegisterWindowException("Unable to register window class '" ~ std.string.toString(wce.lpszClassName) ~ "'");
	}
	
	
	static void registerClass(WNDCLASSEX* wce, WndClassFlags flags) //-flags- indicates which members of -wce- are to be used, otherwise window defaults are used
	{
		if(!(flags & WndClassFlags.STYLE))
			wce.style = CS_OWNDC | CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
		if(!(flags & WndClassFlags.ICONS))
			wce.hIcon = wce.hIconSm = LoadIcon(null, IDI_APPLICATION);
		if(!(flags & WndClassFlags.CURSOR))
			wce.hCursor = LoadCursor(null, IDC_ARROW);
		if(!(flags & WndClassFlags.BACKGROUND))
			wce.hbrBackground = cast(HBRUSH)(COLOR_BTNFACE + 1);
		if(!(flags & WndClassFlags.MENU))
			wce.lpszMenuName = null;
		
		registerClass(wce);
	}
	
	
	void create(DWORD exStyle, char[] className, char[] windowName, DWORD style, int x, int y, int width, int height, HWND hwParent, HMENU hmenu, HINSTANCE hinstance)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
		
		WNDCLASS wc;
		assert(GetClassInfo(hinstance, toStringz(className), &wc) != false);
		assert(wc.lpfnWndProc == (WNDPROC)&allWindowProc);
	}
	body
	{
		addThis();
		
		_hwnd = CreateWindowEx(exStyle, toStringz(className), toStringz(windowName), style, x, y, width, height, hwParent, hmenu, hinstance, (Window*)this);
		if(!_hwnd)
		{
			remove(findThisIndex()); //find it because the order could have been changed during creation
			throw new WindowException("Unable to create window");
		}
	}
	
	
	void bind(HWND hwnd) //just bind this window to this object, no message handling
	{
		_hwnd = hwnd;
	}
	
	
	void unbind()
	{
		_hwnd = null;
		_prevWindowProc = null;
		try
		{
			removeThis();
		}
		catch(WindowException we)
		{
		}
	}
	
	
	LRESULT windowProc(UINT msg, WPARAM wparam, LPARAM lparam) //should be overridden and called and returned from derived class if the message isn't processed: return super.windowProc(msg, wparam, lparam);
	{
		return DefWindowProc(_hwnd, msg, wparam, lparam);
	}
	
	
	//common window functions
	
	void text(char[] newText)
	{
		windowText(_hwnd, newText);
	}
	
	
	int textLength()
	{
		return windowTextLength(_hwnd);
	}
	
	
	char[] text()
	{
		return windowText(_hwnd);
	}
	
	
	void destroy()
	{
		DestroyWindow(_hwnd);
	}
	
	
	void show(int nCmdShow)
	{
		ShowWindow(_hwnd, nCmdShow);
	}
	
	
	void show()
	{
		show(SW_SHOW);
	}
	
	
	void showModal(int nShow)
	{
		.showModal(_hwnd, nShow);
	}
	
	
	void showModal()
	{
		showModal(SW_SHOW);
	}
	
	
	void hide()
	{
		show(SW_HIDE);
	}
	
	
	void minimize()
	{
		//CloseWindow(_hwnd);
		show(SW_MINIMIZE);
	}
	
	
	void restore()
	{
		//OpenIcon(_hwnd);
		show(SW_RESTORE);
	}
	
	
	void maximize()
	{
		show(SW_MAXIMIZE);
	}
	
	
	void display()
	{
		displayWindow(_hwnd);
	}
	
	
	HWND item(int ctrlId)
	{
		return GetDlgItem(_hwnd, ctrlId);
	}
	
	
	void enable(BOOL doEnable)
	{
		EnableWindow(_hwnd, doEnable);
	}
	
	
	void enable()
	{
		enable(true);
	}
	
	
	void disable()
	{
		enable(false);
	}
	
	
	DWORD style()
	{
		return (DWORD)GetWindowLong(_hwnd, GWL_STYLE);
	}
	
	
	void style(DWORD newStyle)
	{
		SetWindowLong(_hwnd, GWL_STYLE, newStyle);
	}
	
	
	DWORD exStyle()
	{
		return (DWORD)GetWindowLong(_hwnd, GWL_EXSTYLE);
	}
	
	
	void exStyle(DWORD newExStyle)
	{
		SetWindowLong(_hwnd, GWL_EXSTYLE, newExStyle);
	}
	
	
	HWND parent()
	{
		return (HWND)GetWindowLong(_hwnd, GWL_HWNDPARENT);
	}
	
	
	void parent(HWND hwNewParent)
	{
		if(!SetParent(_hwnd, hwNewParent))
			throw new WindowException("Unable to set window parent");
	}
	
	
	int id() //control id
	{
		return (int)GetWindowLong(_hwnd, GWL_ID);
	}
	
	
	deprecated alias windowRect rect;
	void windowRect(RECT* r)
	{
		GetWindowRect(_hwnd, r);
	}
	
	
	void clientRect(RECT* cr)
	{
		GetClientRect(_hwnd, cr);
	}
	
	
	void parentArea(RECT* r)
	{
		dwindows.parentArea(_hwnd, r);
	}
	
	
	void center()
	{
		centerWindow(_hwnd);
	}
	
	
	void alignment(Align a)
	{
		dwindows.alignWindow(_hwnd, a);
	}
	
	
	HICON icon(DWORD type) //-type- is ICON_BIG or ICON_SMALL
	{
		return (HICON)SendMessage(_hwnd, WM_GETICON, (WPARAM)type, 0);
	}
	
	
	HICON icon(DWORD type, HICON newIcon) //returns previous icon or null if there wasn't one
	{
		return (HICON)SendMessage(_hwnd, WM_SETICON, (WPARAM)type, (LPARAM)newIcon);
	}
	
	
	HFONT font() //returns null for system font
	{
		return (HFONT)SendMessage(_hwnd, WM_GETFONT, 0, 0);
	}
	
	
	void font(HFONT newFont, BOOL redraw)
	{
		SendMessage(_hwnd, WM_SETFONT, (WPARAM)newFont, MAKELPARAM((WORD)redraw, 0));
	}
	
	
	void redraw(BOOL allow)
	{
		SendMessage(_hwnd, WM_SETREDRAW, (WPARAM)allow, 0);
	}
	
	
	BOOL invalidate(RECT* rect, BOOL eraseBackground)
	{
		return InvalidateRect(_hwnd, rect, eraseBackground);
	}
	
	
	BOOL invalidate(HRGN hrgn, BOOL eraseBackground)
	{
		return InvalidateRgn(_hwnd, hrgn, eraseBackground);
	}
	
	
	BOOL invalidate(BOOL eraseBackground)
	{
		return InvalidateRect(_hwnd, null, eraseBackground);
	}
	
	
	BOOL validate(RECT* rect)
	{
		return ValidateRect(_hwnd, rect);
	}
	
	
	BOOL validate(HRGN hrgn)
	{
		return ValidateRgn(_hwnd, hrgn);
	}
	
	
	BOOL validate()
	{
		return ValidateRect(_hwnd, null);
	}
}


/+
class Dialog: Window
{
	static void registerClass(WNDCLASSEX* wce)
	{
		wce.lpfnWndProc = cast(WNDPROC)&allWindowProc;
		wce.cbSize = WNDCLASSEX.size;
		wce.cbClsExtra = 0;
		wce.cbWndExtra = DLGWINDOWEXTRA;
		
		if(!RegisterClassEx(wce))
			throw new RegisterWindowException("Unable to register dialog class '" ~ std.string.toString(wce.lpszClassName) ~ "'");
	}
	
	
	static void registerClass(WNDCLASSEX* wce, WndClassFlags flags)
	{
		if(!(flags & WndClassFlags.STYLE))
			wce.style = CS_OWNDC | CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
		if(!(flags & WndClassFlags.ICONS))
			wce.hIcon = wce.hIconSm = LoadIcon(null, IDI_APPLICATION);
		if(!(flags & WndClassFlags.CURSOR))
			wce.hCursor = LoadCursor(null, IDC_ARROW);
		if(!(flags & WndClassFlags.BACKGROUND))
			wce.hbrBackground = cast(HBRUSH)(COLOR_BTNFACE + 1);
		if(!(flags & WndClassFlags.MENU))
			wce.lpszMenuName = null;
		
		registerClass(wce);
	}
	
	
	override LRESULT windowProc(UINT msg, WPARAM wparam, LPARAM lparam)
	{
		return DefDlgProc(_hwnd, msg, wparam, lparam);
	}
}
+/


class WindowSubClass: Window
{
	LRESULT windowProc(UINT msg, WPARAM wparam, LPARAM lparam)
	{
		return prevWindowProc(msg, wparam, lparam);
	}
}


class ButtonException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class Button: WindowSubClass
{
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "BUTTON", toStringz(text), style, x, y, width, height, hwParent, (HMENU)ctrlId, null, (Window*)this);
		if(!_hwnd)
			throw new ButtonException("Unable to create button");
	}
	
	
	//common button functions
	
	void click() //simulate click event
	{
		SendMessage(_hwnd, BM_CLICK, 0, 0);
	}
	
	
	DWORD check() //get check state: BST_CHECKED, BST_INTERMEDIATE or BST_UNCHECKED
	{
		return SendMessage(_hwnd, BM_GETCHECK, 0, 0);
	}
	
	
	void check(DWORD newCheckState)
	{
		SendMessage(_hwnd, BM_SETCHECK, (WPARAM)newCheckState, 0);
	}
	
	
	void image(DWORD type, HANDLE image) //-type- is either IMAGE_BITMAP or IMAGE_ICON
	{
		SendMessage(_hwnd, BM_SETIMAGE, (WPARAM)type, (LPARAM)image);
	}
	
	
	HANDLE image(DWORD type) //gets the image, -type- same as when setting
	{
		return cast(HANDLE)SendMessage(_hwnd, BM_GETIMAGE, (WPARAM)type, 0);
	}
	
	
	void state(BOOL highlight)
	{
		SendMessage(_hwnd, BM_SETSTATE, (WPARAM)highlight, 0);
	}
	
	
	void buttonStyle(DWORD style, BOOL redraw) //to not confuse with Window.style
	{
		SendMessage(_hwnd, BM_SETSTYLE, (WPARAM)style, MAKELPARAM((WORD)redraw, 0));
	}
}


class EditException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class Edit: WindowSubClass
{
	const char[] ERROR = "Edit box error";
	const char[] OVERFLOW = "Edit box overflow";
	
	
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "EDIT", toStringz(text), style, x, y, width, height, hwParent, (HMENU)ctrlId, null, (Window*)this);
		if(!_hwnd)
			throw new EditException("Unable to create edit box");
	}
	
	
	//common edit functions
	
	BOOL canUndo()
	{
		return cast(BOOL)SendMessage(_hwnd, EM_CANUNDO, 0, 0);
	}
	
	
	int charFromPos(short x, short y) //client coordinates, 16 bit values only; returns a character index at the specified coordinates, or -1
	{
		return (int)SendMessage(_hwnd, EM_CHARFROMPOS, 0, MAKELPARAM(x, y));
	}
	
	
	void emptyUndoBuffer()
	{
		SendMessage(_hwnd, EM_EMPTYUNDOBUFFER, 0, 0);
	}
	
	
	int limit()
	{
		return SendMessage(_hwnd, EM_GETLIMITTEXT, 0, 0);
	}
	
	
	void limit(int newLimit)
	{
		SendMessage(_hwnd, EM_SETLIMITTEXT, (WPARAM)newLimit, 0);
	}
	
	
	int lineFromChar(int charIndex)
	{
		return (uint)SendMessage(_hwnd, EM_LINEFROMCHAR, (WPARAM)charIndex, 0);
	}
	
	
	int lineIndex(int iline) //returns character index of first character on line -iline- or -1 if the line index does not exist
	{
		int result = cast(int)SendMessage(_hwnd, EM_LINEINDEX, (WPARAM)iline, 0);
		if(result == -1)
			throw new EditException(ERROR);
		return result;
	}
	
	
	int lineLength(int charIndex)
	{
		return (int)SendMessage(_hwnd, EM_LINELENGTH, cast(WPARAM)charIndex, 0);
	}
	
	
	char[] line(int iline)
	{
		int num;
		char[] s;
		
		num = lineIndex(iline); //throws exception if doesn't exist
		num = lineLength(num);
		if(!num)
			return "";
		
		s = new char[num + short.size]; //using short.size since the short must fit and plus it'll be used for the null terminator
		*((short*)((char*)s)) = (short)num;
		num = SendMessage(_hwnd, EM_GETLINE, cast(WPARAM)iline, cast(LPARAM)cast(char*)s);
		return s[0 .. num];
	}
	
	
	void posFromChar(POINT* destPoint, int charIndex)
	{
		SendMessage(_hwnd, EM_POSFROMCHAR, (WPARAM)destPoint, (LPARAM)charIndex);
	}
	
	
	void replaceSel(BOOL canUndo, char[] text)
	{
		SendMessage(_hwnd, EM_REPLACESEL, (WPARAM)canUndo, (LPARAM)toStringz(text));
	}
	
	
	void modify(BOOL modified)
	{
		SendMessage(_hwnd, EM_SETMODIFY, (WPARAM)modified, 0);
	}
	
	
	BOOL modify()
	{
		return cast(BOOL)SendMessage(_hwnd, EM_GETMODIFY, 0, 0);
	}
	
	
	void setSel(int start, int end)
	{
		SendMessage(_hwnd, EM_SETSEL, (WPARAM)start, (LPARAM)end);
	}
	
	
	int getSel(int* start, int* end)
	{
		int result = cast(int)SendMessage(_hwnd, EM_GETSEL, (WPARAM)start, (LPARAM)end);
		if(result == -1)
			throw new EditException(OVERFLOW);
		return result;
	}
	
	
	char[] selText()
	{
		int start, end;
		char[] result;
		getSel(&start, &end);
		end = end - start;
		if(end)
		{
			result = new char[end + 1];
			result = result[0 .. SendMessage(_hwnd, EM_GETSELTEXT, 0, cast(LPARAM)cast(char*)result)];
		}
		return result;
	}
}


class ListBoxException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class ListBox: WindowSubClass
{
	const char[] ERROR = "List box error";
	const char[] INVALID_INDEX = "Invalid index in list box";
	const char[] ERRSPACE = "Insufficient space for list box";
	//const char[] INVALID_OPERATION = "Invalid list box operation";
	
	
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "LISTBOX", toStringz(text), style, x, y, width, height, hwParent, (HMENU)ctrlId, null, (Window*)this);
		if(!_hwnd)
			throw new ListBoxException("Unable to create list box");
	}
	
	
	int addFile(char[] filename)
	{
		LRESULT result = SendMessage(hwnd(), LB_ADDFILE, 0, cast(LPARAM)toStringz(filename));
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(int)result;
	}
	
	
	int dir(UINT attrs, char[] fileSpec)
	{
		LRESULT result = SendMessage(hwnd(), LB_DIR, cast(WPARAM)attrs, cast(LPARAM)toStringz(fileSpec));
		switch(result)
		{
			case LB_ERR:
				throw new ListBoxException(ERROR);
			
			case LB_ERRSPACE:
				throw new ListBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	int addString(char[] text)
	{
		LRESULT result = SendMessage(hwnd(), LB_ADDSTRING, 0, cast(LPARAM)toStringz(text));
		switch(result)
		{
			case LB_ERR:
				throw new ListBoxException(ERROR);
			
			case LB_ERRSPACE:
				throw new ListBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	int deleteString(int index)
	{
		LRESULT result = SendMessage(hwnd(), LB_DELETESTRING, cast(WPARAM)index, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(int)result;
	}
	
	
	int count()
	{
		LRESULT result = SendMessage(hwnd(), LB_GETCOUNT, 0, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(int)result;
	}
	
	
	LRESULT curSel() //returns LB_ERR if no selection
	{
		return SendMessage(hwnd(), LB_GETCURSEL, 0, 0);
	}
	
	
	LRESULT curSel(int index) //if an error occurs, the return value is LB_ERR; if the index parameter is -1, the return value is LB_ERR even though no error occurred
	{
		return SendMessage(hwnd(), LB_SETCURSEL, (WPARAM)index, 0);
	}
	
	
	int getSel(int index)
	{
		LRESULT result = SendMessage(hwnd(), LB_GETSEL, index, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(int)result;
	}
	
	
	void setSel(BOOL selected, int index)
	{
		if(SendMessage(hwnd(), LB_SETSEL, cast(WPARAM)selected, cast(WPARAM)index) == LB_ERR)
			throw new ListBoxException(ERROR);
	}
	
	
	int insertString(int index, char[] text)
	{
		LRESULT result = SendMessage(hwnd(), LB_INSERTSTRING, cast(WPARAM)index, cast(LPARAM)toStringz(text));
		switch(result)
		{
			case LB_ERR:
				throw new ListBoxException(ERROR);
			
			case LB_ERRSPACE:
				throw new ListBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	override char[] text()
	{
		uint len = SendMessage(hwnd(), WM_GETTEXTLENGTH, 0, 0);
		
		if(!len)
			return null;
		
		char[] s = new char[++len];
		SendMessage(hwnd(), WM_GETTEXT, len, cast(LPARAM)cast(char*)s);
		return s[0 .. len - 1];
	}
	
	
	override void text(char[] newText)
	{
		SendMessage(hwnd(), WM_SETTEXT, 0, cast(LPARAM)toStringz(newText));
	}
	
	
	LPARAM itemData(int index)
	{
		LRESULT result = SendMessage(hwnd(), LB_GETITEMDATA, cast(WPARAM)index, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(LPARAM)result;
	}
	
	
	int itemTextLen(int index)
	{
		LRESULT result = SendMessage(hwnd(), LB_GETTEXTLEN, cast(WPARAM)index, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(int)result;
	}
	
	
	char[] itemText(int index)
	{
		char[] s = new char[itemTextLen(index) + 1];
		int len = cast(int)SendMessage(hwnd(), LB_GETTEXT, cast(WPARAM)index, cast(LPARAM)cast(char*)s); //no need to test for LB_ERR since itemTextLen did
		return s[0 .. len];
	}
	
	
	int topIndex()
	{
		return cast(int)SendMessage(hwnd(), LB_GETTOPINDEX, 0, 0);
	}
	
	
	int initStorage(int cItems, DWORD cb)
	{
		LRESULT result = SendMessage(hwnd(), LB_INITSTORAGE, cast(WPARAM)cItems, cast(LPARAM)cb);
		if(result == LB_ERRSPACE)
			throw new ListBoxException(ERRSPACE);
		return cast(int)result;
	}
	
	
	LRESULT itemFromPoint(short x, short y) //return value contains the index of the nearest item in the low-order word; the high-order word is zero if the specified point is in the client area of the list box, or one if it is outside the client area
	{
		return SendMessage(hwnd(), LB_ITEMFROMPOINT, 0, MAKELPARAM(x, y));
	}
	
	
	void reset()
	{
		SendMessage(hwnd(), LB_RESETCONTENT, 0, 0);
	}
	
	
	LRESULT selectString(int indexStart, char[] str) //returns LB_ERR if the search was unsuccessful
	{
		return SendMessage(hwnd(), LB_SELECTSTRING, cast(WPARAM)indexStart, cast(LPARAM)toStringz(str));
	}
	
	
	void horizontalExtent(int cxExtent)
	{
		SendMessage(hwnd(), LB_SETHORIZONTALEXTENT, cast(WPARAM)cxExtent, 0);
	}
	
	
	void itemData(int index, DWORD dwData)
	{
		if(SendMessage(hwnd(), LB_SETITEMDATA, cast(WPARAM)index, cast(LPARAM)dwData) == LB_ERR)
			throw new ListBoxException(ERROR);
	}
	
	
	void itemHeight(int index, int height)
	{
		if(SendMessage(hwnd(), LB_SETITEMHEIGHT, cast(WPARAM)index, MAKELPARAM(height, 0)) == LB_ERR)
			throw new ListBoxException(ERROR);
	}
	
	
	WORD locale(WORD localeId) //returns previous locale
	{
		LRESULT result = SendMessage(hwnd(), LB_SETLOCALE, cast(WPARAM)localeId, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR); //"Invalid locale"
		return cast(WORD)result;
	}
	
	
	LRESULT topIndex(int index) //returns LB_ERR on failure
	{
		return SendMessage(hwnd(), LB_SETTOPINDEX, cast(WPARAM)index, 0);
	}
	
	
	uint selCount()
	{
		LRESULT result = SendMessage(hwnd(), LB_GETSELCOUNT, 0, 0);
		if(result == LB_ERR)
			throw new ListBoxException(ERROR);
		return cast(uint)result;
	}
	
	
	int[] selItems()
	{
		uint len = selCount(); //throws exception on err
		int[] result;
		if(len)
		{
			result = new int[len];
			SendMessage(hwnd(), LB_GETSELITEMS, result.length, cast(LPARAM)cast(int*)result);
		}
		return result;
	}
}


class TabControlException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class TabControl: WindowSubClass //InitCommonControls() must have been called
{
	const char[] ERROR = "Tab control error";
	
	
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "SysTabControl32", toStringz(text), style, x, y, width, height, hwParent, (HMENU)ctrlId, null, (Window*)this);
		if(!_hwnd)
			throw new TabControlException("Unable to create tab control");
	}
	
	
	void adjustRect(BOOL fLarger, RECT* prc)
	{
		TabCtrl_AdjustRect(hwnd(), fLarger, prc);
	}
	
	
	void deleteAllItems()
	{
		if(!TabCtrl_DeleteAllItems(hwnd())) //does this ever fail?
			throw new TabControlException(ERROR);
	}
	
	
	void deleteItem(int iItem)
	{
		if(!TabCtrl_DeleteItem(hwnd(), iItem))
			throw new TabControlException(ERROR);
	}
	
	
	int curFocus()
	{
		return TabCtrl_GetCurFocus(hwnd());
	}
	
	
	int curSel() //-1 is returned if no selection
	{
		return TabCtrl_GetCurSel(hwnd());
	}
	
	
	HIMAGELIST imageList()
	{
		return TabCtrl_GetImageList(hwnd());
	}
	
	
	void getItem(int iItem, TC_ITEM* pitem)
	{
		if(!TabCtrl_GetItem(hwnd(), iItem, pitem))
			throw new TabControlException(ERROR);
	}
	
	
	int itemCount()
	{
		return TabCtrl_GetItemCount(hwnd());
	}
	
	
	void itemRect(int iItem, RECT* prc)
	{
		TabCtrl_GetItemRect(hwnd(), iItem, prc); //returns TRUE/FALSE but never fails?
	}
	
	
	int rowCount()
	{
		return TabCtrl_GetRowCount(hwnd());
	}
	
	
	HWND toolTips() //returns the handle to the tooltip control if successful or null otherwise
	{
		return cast(HWND)TabCtrl_GetToolTips(hwnd());
	}
	
	
	int hitTest(TC_HITTESTINFO* pinfo) //returns the index of the tab or  -1 if no tab is at the specified position
	{
		return TabCtrl_HitTest(hwnd(), pinfo);
	}
	
	
	int insertItem(int iItem, TC_ITEM* pitem) //returns the new tab index
	{
		int result = cast(int)TabCtrl_InsertItem(hwnd(), iItem, pitem);
		if(result == -1)
			throw new TabControlException(ERROR);
		return result;
	}
	
	
	void removeImage(int iImage)
	{
		TabCtrl_RemoveImage(hwnd(), iImage);
	}
	
	
	void curFocus(int iItem)
	{
		TabCtrl_SetCurFocus(hwnd(), iItem);
	}
	
	
	int curSel(int iItem) //returns the index of the previously selected tab (using this does not cause selection notifications to be sent)
	{
		int result = cast(int)TabCtrl_SetCurSel(hwnd(), iItem);
		if(result == -1)
			throw new TabControlException(ERROR);
		return result;
	}
	
	
	HIMAGELIST imageList(HIMAGELIST himl) //returns the handle to the previous image list or NULL if there is no previous image list
	{
		return cast(HIMAGELIST)TabCtrl_SetImageList(hwnd(), himl);
	}
	
	
	void setItem(int iItem, TC_ITEM* pitem)
	{
		if(!TabCtrl_SetItem(hwnd(), iItem, pitem))
			throw new TabControlException(ERROR);
	}
	
	
	void itemExtra(int cb)
	{
		if(!TabCtrl_SetItemExtra(hwnd(), cb))
			throw new TabControlException(ERROR);
	}
	
	
	DWORD itemSize(int cx, int cy)
	{
		return TabCtrl_SetItemSize(hwnd(), cx, cy);
	}
	
	
	void padding(int cx, int cy)
	{
		TabCtrl_SetPadding(hwnd(), cx, cy);
	}
	
	
	void toolTips(HWND hwndTT)
	{
		TabCtrl_SetToolTips(hwnd(), hwndTT);
	}
}


class StaticException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class Static: WindowSubClass
{
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "STATIC", toStringz(text), style, x, y, width, height, hwParent, cast(HMENU)ctrlId, null, cast(Window*)this);
		if(!_hwnd)
			throw new StaticException("Unable to create static");
	}
	
	
	void image(DWORD type, HANDLE image) //-type- is either IMAGE_BITMAP, IMAGE_CURSOR, IMAGE_ENHMETAFILE or IMAGE_ICON
	{
		SendMessage(_hwnd, STM_SETIMAGE, cast(WPARAM)type, cast(LPARAM)image);
	}
	
	
	HANDLE image(DWORD type) //gets the image, same -type- as set
	{
		return cast(HANDLE)SendMessage(_hwnd, STM_GETIMAGE, cast(WPARAM)type, 0);
	}
	
	
	HICON icon(HICON hicon) //associate an icon with an icon control, returns previous or null on error
	{
		return cast(HICON)SendMessage(_hwnd, STM_SETICON, cast(WPARAM)hicon, 0);
	}
	
	
	HICON icon() //retrieve the handle of the icon associated with a static control that has the SS_ICON style
	{
		return cast(HICON)SendMessage(_hwnd, STM_GETICON, 0, 0);
	}
}


class ComboBoxException: WindowException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class ComboBox: WindowSubClass
{
	const char[] ERROR = "Combo box error";
	const char[] INVALID_INDEX = "Invalid index in combo box";
	const char[] ERRSPACE = "Insufficient space for combo box";
	const char[] INVALID_OPERATION = "Invalid combo box operation";
	
	
	//remember height is the height of the default drop box, not the height of the static/editbox
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		_hwnd = CreateWindowEx(exStyle, "COMBOBOX", toStringz(text), style, x, y, width, height, hwParent, cast(HMENU)ctrlId, null, cast(Window*)this);
		if(!_hwnd)
			throw new ComboBoxException("Unable to create combo box");
	}
	
	
	int addString(char[] str)
	{
		LRESULT result = SendMessage(hwnd(), CB_ADDSTRING, 0, cast(LPARAM)toStringz(str));
		switch(result)
		{
			case CB_ERR:
				throw new ComboBoxException(ERROR);
			
			case CB_ERRSPACE:
				throw new ComboBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	int deleteString(int index)
	{
		LRESULT result = SendMessage(hwnd(), CB_DELETESTRING, cast(WPARAM)index, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(INVALID_INDEX);
		return cast(int)result;
	}
	
	
	int dir(UINT attrs, char[] fileSpec)
	{
		LRESULT result = SendMessage(hwnd(), CB_DIR, cast(WPARAM)attrs, cast(LPARAM)toStringz(fileSpec));
		switch(result)
		{
			case CB_ERR:
				throw new ComboBoxException(ERROR);
			
			case CB_ERRSPACE:
				throw new ComboBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	LRESULT findString(int indexStart, char[] find) //returns CB_ERR if not found
	{
		return SendMessage(hwnd(), CB_FINDSTRING, cast(WPARAM)indexStart, cast(LPARAM)toStringz(find));
	}
	
	
	LRESULT findStringExact(int indexStart, char[] find) //returns CB_ERR if not found
	{
		return SendMessage(hwnd(), CB_FINDSTRINGEXACT, cast(WPARAM)indexStart, cast(LPARAM)toStringz(find));
	}
	
	
	int count()
	{
		LRESULT result = SendMessage(hwnd(), CB_GETCOUNT, 0, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	LRESULT curSel() //returns CB_ERR if no selection
	{
		return SendMessage(hwnd(), CB_GETCURSEL, 0, 0);
	}
	
	
	void getDroppedControlRect(RECT* rect)
	{
		SendMessage(hwnd(), CB_GETDROPPEDCONTROLRECT, 0, cast(LPARAM)rect);
	}
	
	
	BOOL droppedState()
	{
		return cast(BOOL)SendMessage(hwnd(), CB_GETDROPPEDSTATE, 0, 0);
	}
	
	
	int droppedWidth()
	{
		LRESULT result = SendMessage(hwnd(), CB_GETDROPPEDWIDTH, 0, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	DWORD getEditSel(DWORD* start, DWORD* end)
	{
		return cast(DWORD)SendMessage(hwnd(), CB_GETEDITSEL, cast(WPARAM)start, cast(LPARAM)end);
	}
	
	
	BOOL extendedUi()
	{
		return cast(BOOL)SendMessage(hwnd(), CB_GETEXTENDEDUI, 0, 0);
	}
	
	
	int horizontalExtent()
	{
		return cast(int)SendMessage(hwnd(), CB_GETHORIZONTALEXTENT, 0, 0);
	}
	
	
	DWORD itemData(int index)
	{
		LRESULT result = SendMessage(hwnd(), CB_GETITEMDATA, cast(WPARAM)index, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(DWORD)result;
	}
	
	
	int itemHeight(int index)
	{
		LRESULT result = SendMessage(hwnd(), CB_GETITEMHEIGHT, cast(WPARAM)index, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	char[] itemText(int index)
	{
		char[] result = new char[itemTextLen(index) + 1];
		int len = cast(int)SendMessage(hwnd(), CB_GETLBTEXT, cast(WPARAM)index, cast(LPARAM)cast(char*)result); //no need to test for CB_ERR since itemTextLen did
		return result[0 .. len];
	}
	
	
	int itemTextLen(int index)
	{
		LRESULT result = SendMessage(hwnd(), CB_GETLBTEXTLEN, cast(WPARAM)index, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(INVALID_INDEX);
		return cast(int)result;
	}
	
	
	DWORD locale()
	{
		return cast(DWORD)SendMessage(hwnd(), CB_GETLOCALE, 0, 0);
	}
	
	
	int topIndex()
	{
		LRESULT result = SendMessage(hwnd(), CB_GETTOPINDEX, 0, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	int initStorage(int cItems, DWORD cb)
	{
		LRESULT result = SendMessage(hwnd(), CB_INITSTORAGE, cast(WPARAM)cItems, cast(LPARAM)cb);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	int insertString(int index, char[] str)
	{
		LRESULT result = SendMessage(hwnd(), CB_INSERTSTRING, cast(WPARAM)index, cast(LPARAM)cast(char*)str);
		switch(result)
		{
			case CB_ERR:
				throw new ComboBoxException(ERROR);
			
			case CB_ERRSPACE:
				throw new ComboBoxException(ERRSPACE);
			
			default:
				return cast(int)result;
		}
	}
	
	
	override char[] text()
	{
		uint len = SendMessage(hwnd(), WM_GETTEXTLENGTH, 0, 0);
		
		if(!len)
			return null;
		
		char[] s = new char[++len];
		SendMessage(hwnd(), WM_GETTEXT, len, cast(LPARAM)cast(char*)s);
		return s[0 .. len - 1];
	}
	
	
	override void text(char[] newText)
	{
		SendMessage(hwnd(), WM_SETTEXT, 0, cast(LPARAM)toStringz(newText));
	}
	
	
	void limit(int newLimit)
	{
		SendMessage(hwnd(), CB_LIMITTEXT, cast(WPARAM)newLimit, 0);
	}
	
	
	void reset()
	{
		SendMessage(hwnd(), CB_RESETCONTENT, 0, 0);
	}
	
	
	LRESULT selectString(int indexStart, char[] str) //returns CB_ERR if the search was unsuccessful
	{
		return SendMessage(hwnd(), CB_SELECTSTRING, cast(WPARAM)indexStart, cast(LPARAM)toStringz(str));
	}
	
	
	LRESULT curSel(int index) //if index is -1 or past the end, returns CB_ERR and clears the selection
	{
		return SendMessage(hwnd(), CB_SETCURSEL, cast(WPARAM)index, 0);
	}
	
	
	int droppedWidth(int width)
	{
		LRESULT result = SendMessage(hwnd(), CB_SETDROPPEDWIDTH, cast(WPARAM)width, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR);
		return cast(int)result;
	}
	
	
	void setEditSel(short start, short end)
	{
		if(SendMessage(hwnd(), CB_SETEDITSEL, 0, MAKELPARAM(start, end)) == CB_ERR)
			throw new ComboBoxException(INVALID_OPERATION);
	}
	
	
	void extendedUi(BOOL fext)
	{
		if(SendMessage(hwnd(), CB_SETEXTENDEDUI, cast(WPARAM)fext, 0) == CB_ERR)
			throw new ComboBoxException(ERROR);
	}
	
	
	void horizontalExtent(int cxExtent)
	{
		SendMessage(hwnd(), CB_SETHORIZONTALEXTENT, cast(WPARAM)cxExtent, 0);
	}
	
	
	void itemData(int index, DWORD dwData)
	{
		if(SendMessage(hwnd(), CB_SETITEMDATA, cast(WPARAM)index, cast(LPARAM)dwData) == CB_ERR)
			throw new ComboBoxException(ERROR);
	}
	
	
	void itemHeight(int index, int height)
	{
		if(SendMessage(hwnd(), CB_SETITEMHEIGHT, cast(WPARAM)index, cast(LPARAM)height) == CB_ERR)
			throw new ComboBoxException(ERROR);
	}
	
	
	WORD locale(WORD localeId) //returns previous locale
	{
		LRESULT result = SendMessage(hwnd(), CB_SETLOCALE, cast(WPARAM)localeId, 0);
		if(result == CB_ERR)
			throw new ComboBoxException(ERROR); //"Invalid locale"
		return cast(WORD)result;
	}
	
	
	LRESULT topIndex(int index) //returns CB_ERR on failure
	{
		return SendMessage(hwnd(), CB_SETTOPINDEX, cast(WPARAM)index, 0);
	}
	
	
	void showDropdown(BOOL fshow)
	{
		SendMessage(hwnd(), CB_SHOWDROPDOWN, cast(WPARAM)fshow, 0);
	}
}


extern(Windows) LRESULT allWindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	int i;
	
	if(msg == WM_NCCREATE)
	{
		i = Window.findIndex((Window)((Window*)(((CREATESTRUCT*)lparam).lpCreateParams)));
		if(i == -1)
			return FALSE;
		
		Window.windows[i]._hwnd = hwnd;
		return Window.windows[i].windowProc(msg, wparam, lparam);
	}
	else
	{
		i = Window.findIndex(hwnd);
		if(i == -1)
			return 0;
		
		LRESULT lresult = Window.windows[i].windowProc(msg, wparam, lparam);
		if(msg == WM_DESTROY)
		{
			i = Window.findIndex(hwnd); //anything can happen in -windowProc-
			if(i != -1)
				Window.windows[i].unbind();
		}
		
		return lresult;
	}
}

