/*
	Copyright (C) 2004 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/


//debug = TABFRAME_UNITTEST;


module tabframe;

private import dwindows;
private import std.string;


private bit isInit = false;
	

class TabFrameException: TabControlException
{
	this(char[] msg)
	{
		super(msg);
	}
}


class TabFrame: TabControl
{
	protected Tab[] _tabs = null;
	
	
	//parent should call this in its WM_NOTIFY message
	//if the return value is nonzero, it must be returned from the window procedure
	static LRESULT notify(WPARAM wparam, LPARAM lparam)
	{
		NMHDR* nmhdr = cast(NMHDR*)lparam;
		Tab tab;
		Window win;
		uint num;
		
		win = Window.find(nmhdr.hwndFrom);
		if(!win || !(cast(TabFrame)win))
			return 0;
		
		with(cast(TabFrame)win)
		{
			switch(nmhdr.code)
			{
				case TCN_SELCHANGING:
					tab = _tabs[curSel()];
					if(tab._lock)
					{
						tab.onLockDeny();
						return cast(LRESULT)TRUE; //prevent selection
					}
					if(tab.onUnselEx())
						return cast(LRESULT)TRUE; //prevent selection
					tab.hide();
					break;
				
				case TCN_SELCHANGE:
					selDisplay(_tabs[curSel()]);
					break;
				
				case TCN_KEYDOWN:
					_tabs[curSel()].onKeyDown((cast(TC_KEYDOWN*)lparam).wVKey, (cast(TC_KEYDOWN*)lparam).flags);
					break;
				
				default: ;
			}
			return 0;
		}
	}
	
	
	this()
	{
		if(!isInit)
		{
			isInit = true;
			
			WNDCLASSEX wce;
			(cast(ubyte*)&wce)[0 .. wce.size] = 0;
			wce.style = CS_PARENTDC | CS_NOCLOSE;
			wce.lpszClassName = Tab.CLASSNAME;
			wce.hInstance = Tab.hinstance = GetModuleHandle(null);
			Window.registerClass(&wce);
		}
	}
	
	
	void create(DWORD exStyle, char[] text, DWORD style, int x, int y, int width, int height, HWND hwParent, int ctrlId)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		HWND hwnd = CreateWindowEx(exStyle, "SysTabControl32", toStringz(text), style, x, y, width, height, hwParent, (HMENU)ctrlId, null, (Window*)this);
		if(!hwnd)
			throw new TabFrameException("Unable to create tab control for tab-frame");
		subClass(hwnd);
	}
	
	
	Tab[] tabs()
	{
		return _tabs;
	}
	
	
	int findTabIndex(Tab tab)
	{
		int i;
		for(i = 0; i != _tabs.length; i++)
		{
			if(_tabs[i] === tab)
				return i;
		}
		return -1;
	}
	
	
	protected void selDisplay(Tab tab)
	{
		RECT r;
		clientRect(&r);
		adjustRect(FALSE, &r);
		SetWindowPos(tab.hwnd(), null, r.left, r.top, r.right - r.left, r.bottom - r.top, SWP_NOACTIVATE | SWP_NOZORDER);
		tab.onSel();
		tab.show();
	}
	
	
	void deleteAllItems()
	{
		super.deleteAllItems(); //throws exception on failure
		
		foreach(Tab tab; _tabs)
		{
			tab.destroy();
		}
		
		_tabs = null;
	}
	
	
	void deleteItem(int iItem)
	{
		super.deleteItem(iItem); //throws exception on failure
		
		_tabs[iItem].destroy();
		_tabs = _tabs[0 .. iItem] ~ _tabs[iItem + 1 .. _tabs.length];
		
		int i = curSel();
		if(i != -1)
			selDisplay(_tabs[i]);
	}
	
	
	int insertItem(int iItem, Tab tab)
	{
		int ibefore = curSel();
		int newSel;
		
		TC_ITEM ti;
		ti.mask = TCIF_TEXT | TCIF_IMAGE | TCIF_PARAM;
		ti.pszText = tab._text;
		ti.iImage = tab._iImage;
		ti.lParam = cast(LPARAM)cast(Tab*)tab;
		int result = super.insertItem(iItem, &ti); //throws exception on failure
		
		tab.create(this);
		
		if(result == _tabs.length)
		{
			_tabs ~= tab;
		}
		else
		{
			Tab[] tt = tabs[0 .. iItem].dup;
			tt.length = tt.length + 1;
			
			tt[tt.length - 1] = tab;
			_tabs = tt ~ _tabs[iItem .. _tabs.length];
		}
		
		newSel = curSel();
		if(newSel != result)
			goto done;
		
		if(ibefore == -1)
			goto set_disp;
		
		if(ibefore > result) //need to add 1 to ibefore
		{
			_tabs[++ibefore].hide();
		}
		else if(ibefore < result)
		{
			_tabs[ibefore].hide();
		}
		
		set_disp:
		selDisplay(_tabs[result]);
		
		done:
		return result;
	}
	
	
	//returns true if it's safe to destroy the window (based on locking)
	bit canDestroy()
	{
		with(_tabs[curSel()])
		{
			return !_lock && !onUnselEx();
		}
	}
}


class Tab: Window
{
	const static char[] CLASSNAME = "TabFrame_Tab";
	
	static HINSTANCE hinstance;
	
	private
	{
		bit _lock = false; //if this tab is visible, other tabs can't be clicked
		char[] _text;
		int _iImage;
		TabFrame tframe;
	}
	
	
	this(char[] text, int iImage)
	{
		_text = toStringz(text)[0 .. text.length];
		_iImage = iImage;
	}
	
	
	this(char[] text)
	{
		this(text, -1);
	}
	
	
	private void create(TabFrame tframe)
	in
	{
		assert(_hwnd == null);
		assert(findThisIndex() == -1);
	}
	body
	{
		this.tframe = tframe;
		try
		{
			super.create(WS_EX_CONTROLPARENT | WS_EX_TRANSPARENT, CLASSNAME, "", WS_CHILD | WS_TABSTOP, 0, 0, 0, 0, tframe.hwnd(), null, hinstance);
		}
		catch(WindowException we)
		{
			throw new TabFrameException("Unable to create tab-frame tab");
		}
	}
	
	
	void text(char[] newText)
	{
		TC_ITEM ti;
		ti.mask = TCIF_TEXT;
		ti.pszText = _text = toStringz(newText)[0 .. newText.length];
		tframe.setItem(tframe.findTabIndex(this), &ti);
	}
	
	
	char[] text()
	{
		return _text;
	}
	
	
	TabFrame frame()
	{
		return tframe;
	}
	
	
	void lock()
	{
		_lock = true;
	}
	
	
	void unlock()
	{
		_lock = false;
	}
	
	
	void onSel() //override to be notified when this tab is about to be selected
	{
	}
	
	
	void onUnsel() //override to be notified when this tab is about to be unselected
	{
	}
	
	
	//override to be notified when this tab is to be unselected (won't call onUnsel if this is overridden)
	//return TRUE to prevent selection change like lock would
	BOOL onUnselEx()
	{
		onUnsel();
		return FALSE;
	}
	
	
	//override to be notified when a selection was denied due to this tab being locked
	//DO NOT attempt to set the selection in this function or it could result in a stack overflow!
	void onLockDeny()
	{
	}
	
	
	void onKeyDown(int virtKey, UINT flags) //override to be notified of key down events
	{
	}
}


debug(TABFRAME_UNITTEST)
unittest
{
	class TestWin: Window
	{
		override LRESULT windowProc(UINT msg, WPARAM wparam, LPARAM lparam)
		{
			LRESULT result;
			
			switch(msg)
			{
				case WM_NOTIFY:
					result = TabFrame.notify(wparam, lparam);
					if(result)
						return result;
					break;
				
				case WM_DESTROY:
					PostQuitMessage(0);
					break;
				
				default:
					return super.windowProc(msg, wparam, lparam);
			}
			return 0;
		}
	}
	
	
	InitCommonControls();	
	
	const char[] UNITTEST_CLASSNAME = "TabFrame_unittest";
	WNDCLASSEX wce;
	wce.hInstance = GetModuleHandle(null);
	wce.lpszClassName = UNITTEST_CLASSNAME;
	Window.registerClass(&wce, WndClassFlags.NONE);
	
	TestWin twin = new TestWin;
	twin.create(0, UNITTEST_CLASSNAME, "TabFrame unittest", WS_OVERLAPPEDWINDOW, 0, 0, 200, 200, null, null, wce.hInstance);
	RECT r;
	twin.clientRect(&r);
	
	with(new TabFrame)
	{
		create(0, "", WS_CHILD | WS_VISIBLE, 0, 0, r.right, r.bottom, twin.hwnd(), 0);
		insertItem(0, new Tab("some tab"));
		insertItem(1, new Tab("weee"));
	}
	
	twin.show();
	
	MSG msg;
	while(GetMessage(&msg, null, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

